/*
 * SienaEventBus.java
 *
 * Created on November 2, 2001, 12:46 PM
 */

package edu.cmu.cs.able.gaugeInfrastructure.Siena;
import edu.cmu.cs.able.gaugeInfrastructure.Events.*;
import edu.cmu.cs.able.gaugeInfrastructure.Siena.Events.*;
import edu.cmu.cs.able.gaugeInfrastructure.util.*;
import edu.cmu.cs.able.gaugeInfrastructure.*;
import siena.*;
import java.util.*;
/** The wrapper around the Siena event bus. This marshals and unmarshals events,
 * provides an announcement interface, as well as a request/response interface.
 * Currently, this class is a little cludgy. In future versions it will be
 * split into four classses - an abstact root class and one for the each type of bus
 * (Gauge, GaugeMgr, GaugeConsumer) that deals with specific events.
 *
 * @author Bradley Schmerl
 * @version 1.1
 */
public class SienaEventBus implements siena.Notifiable {

    
    class NotificationHandlerGauge {
        public Notification filterAsNotification;
        public Filter filter;
        public Object handler;
        public GaugeControl gauge;
    }
    
    /** The tag for a ConfiguredEvent.
     */    
    protected final static String CONFIGURED = "gaugeConfigured";
    /** The tag for a DeletedEvent.
     */    
    protected final static String DELETED = "gaugeDeleted";
    /** The tag for a CreatedEvent.
     */    
    protected final static String CREATED = "gaugeCreated";
    /** The tag for a {@link MultipleValueEvent}
     */    
    protected final static String MULTIPLEVALUES = "gaugeMV";
    /** The tag for a {@link ValueEvent}
     */    
    protected final static String VALUE = "onReport";
    /** The tag for a {@link DeleteEvent}
     */    
    protected final static String DELETE = "deleteGauge";
    /** The tag for a {@link ConfigureEvent}
     */    
    protected final static String CONFIGURE = "configureGauge";
    /** The tag for a {@link CreateEvent}
     */    
    protected final static String CREATE = "createGauge";
    /** The tag for a {@link QueryMetaInfoEvent}
     */    
    protected final static String METAINFO = "queryGaugeMetaInfo";
    /** The tag for a {@link QueryValueEvent}
     */    
    protected final static String QUERYVALUE = "queryValue";
    /** The tag for a {@link QueryAllValuesEvent}
     */    
    protected final static String QUERYALLVALUES = "queryAllValues";
    /** The tag for a {@link QueryStateEvent}
     */    
    protected final static String QUERYSTATE = "queryGaugeState";
    /** The tag for a {@link FindGaugeMgrEvent}
     */    
    protected final static String FINDGMGR = "findGaugeMgr";
    /** The tag for event direction=request
     */    
    protected final static String REQUEST = "request";
    /** The tag for event direction=announce
     */    
    protected final static String ANNOUNCE = "announce";
    /** The tag for event direction=response
     */    
    protected final static String RESPONSE = "reply";
    /** The tag for {@link GaugeInitEvent}
     */    
    protected final static String GAUGEINIT = "gaugeInit";
    /** The tag for {@link GMgrDeleteGaugeEvent}
     */    
    protected final static String GMGRDELETEGAUGE = "GMgrDeleteGauge";
    /** The tag for a CreationFinishedEvent
     */
    protected final static String FINISHEDCREATION = "FinishedCreation";
    
    /** The tag used in a Siena notification to denote the event type attribute name
     */    
    protected final static String EVENTTYPE = "eventType";
    /** The tag used in a Siena notification to denote the gauge manager ID attribute name
     */    
    protected final static String GAUGEMGRID = "gaugeMgr";
    /** The tag used in a Siena notification to denote the setup attribute name
     */    
    protected final static String SETUP = "setup";
    /** The tag used in a Siena notification to denote the mappings attribute name
     */    
    protected final static String MAPPINGS = "mapping";
    /** The tag used in a Siena notification to denote the config attribute name
     */    
    protected final static String CONFIG = "config";
    /** The tag used in a Siena notification to denote the status attribute name
     */    
    protected final static String STATUS = "status";
    /** The tag used in a Siena notification to denote the initiator attribute name
     */    
    protected final static String INITIATOR = "initiator";
    /** The tag used in a Siena notification to denote the event ID attribute name
     */    
    protected final static String EVENTID = "eventID";
    /** The tag used in a Siena notification to denote the direction attribute name
     */    
    protected final static String DIRECTION = "direction";
    
    /** How long to wait for a blocking response (in milliseconds)
     */    
    protected final static int TIMEOUT = 30000;
    /** How many times to retry a blocking wait
     */    
    protected final static int RETRIES = 3;
    
    /** The siena bus
     */    
    public static HierarchicalDispatcher siena = new HierarchicalDispatcher();
    /** Only one event bus can be constructed. This is it.
     */    
    public static SienaEventBus theEventBus = null;
    
    private static boolean initialized = false;
    private String id = null;
    private Vector eventBindings = new Vector ();
    private Vector gauges = new Vector ();
    
    /** The Siena port
     */    
    public static String senp;
    
    /** Creates a new event bus
     * @param uri The Siena port to connect to
     */    
    public SienaEventBus(String uri) {
        try {
            senp = uri;
            siena.setMaster (senp);
        }
        catch (Exception ex) {
            ex.printStackTrace ();
        }
        id = Global.generateUniqueID ();
        theEventBus = this;
        initialized = true;
    }
    
    /** Publishes an announcement on the bus
     * @param event The event to be announced
     * @return Whether the announcement was publishes successfully
     */    
    public boolean publishAnnouncement (GaugeEvent event) {
        Notification not = null;
        // Work out the right event, and pack it appropriately
        if (event instanceof ConfiguredEvent) {
            not = packConfiguredEvent ((ConfiguredEvent )event);
        }
        else if (event instanceof CreatedEvent) {
            not = packCreatedEvent ((CreatedEvent )event);
        }
        else if (event instanceof DeletedEvent) {
            not = packDeletedEvent ((DeletedEvent )event);
        }
        else if (event instanceof MultipleValueEvent) {
            not = packMultipleValueEvent ((MultipleValueEvent )event);
        }
        else if (event instanceof ValueEvent) {
            not = packValueEvent ((ValueEvent )event);
        }
        else if (event instanceof CreationFinishedEvent) {
            not = packCreationFinishedRequest ((CreationFinishedEvent )event);
        }
        
        // This event is an announcement - no reply is expected
        not.putAttribute (DIRECTION, ANNOUNCE);
        
        // publish the event on Siena
        try {
            Global.debug (not.toString ());
            siena.publish (not);
        }
        catch (SienaException ex) {
            ex.printStackTrace ();
            return false;
        }
        return true;
    }
    
    /** Publishes a request on the bus. Gauge consumers or managers
     * call this when they wish to communicate directly to a gauge or
     * gauge manager. The event must have the appropriate request side
     * attributes filled in.
     * @param event The request to publish
     * @return Whether the event was published successfully
     */    
    public GaugeEvent publishRequest(GaugeEvent event) {
        Notification not = null;

        // Determine the right event and pack it into a notification
        if (event instanceof QueryMetaInfoEvent) {
            not = packQueryMetaInfoEventRequest ((QueryMetaInfoEvent )event);
        }
        else if (event instanceof ConfigureGaugeEvent) {
            not = packConfigureEventRequest ((ConfigureGaugeEvent )event);
        }
        else if (event instanceof CreateGaugeEvent) {
            not = packCreateEventRequest ((CreateGaugeEvent )event);
        }
        else if (event instanceof DeleteGaugeEvent) {
            not = packDeleteEventRequest ((DeleteGaugeEvent )event);
        }
        else if (event instanceof QueryValueEvent) {
            not = packQueryValueEventRequest ((QueryValueEvent )event);
        }
        else if (event instanceof QueryAllValuesEvent) {
            not = packQueryAllValuesEventRequest ((QueryAllValuesEvent )event);
        }
        else if (event instanceof QueryGaugeStateEvent) {
            not = packQueryGaugeStateRequest ((QueryGaugeStateEvent )event);
        }
        else if (event instanceof FindGaugeMgrEvent) {
            not = packFindGaugeMgrRequest ((FindGaugeMgrEvent )event);
        }
        else if (event instanceof GaugeInitEvent) {
            not = packGaugeInitRequest ((GaugeInitEvent )event);
        }
        else if (event instanceof GMgrDeleteGaugeEvent) {
            not = packGMgrDeleteGaugeRequest ((GMgrDeleteGaugeEvent )event);
        }

        
        // This event is a request, so we need to add who initiated the event
        // so that replies come back to the right place
        not.putAttribute (DIRECTION, REQUEST);
        not.putAttribute (INITIATOR, id);
        
        // Prepare for response
        Filter f = new Filter ();
        NotificationBuffer nb = prepareForReply (event,f);
        
        // Publish the event using Siena
        try {
            Global.debug (not.toString ());
            siena.publish (not);
        }
        catch (SienaException ex) {
            ex.printStackTrace ();
            event.status = false;
            return event;
        }
        GaugeEvent reply = waitForReply (event, nb, f);
        return reply;
    }
    
    /** Publishes a response. The event must have the appropriate reply
     * side attributes filled in. This is used to reply to a request
     * event.
     * @param event The reply to be published
     * @param initiator Who the reply should go to
     * @return Whether the reply was published correctly
     */    
    public boolean publishResponse (GaugeEvent event, String initiator) {
        Notification not = null;
        
        // Find the right event, and pack the reply into a notification
        if (event instanceof QueryMetaInfoEvent) {
            not = this.packQueryMetaInfoEventReply ((QueryMetaInfoEvent )event);
        }
        else if (event instanceof ConfigureGaugeEvent) {
            not = packConfigureEventReply ((ConfigureGaugeEvent )event);
        }
        else if (event instanceof CreateGaugeEvent) {
            not = packCreateEventReply ((CreateGaugeEvent )event);
        }
        else if (event instanceof DeleteGaugeEvent) {
            not = packDeleteEventReply ((DeleteGaugeEvent )event);
        }
        else if (event instanceof QueryValueEvent) {
            not = packQueryValueEventReply ((QueryValueEvent )event);
        }
        else if (event instanceof QueryAllValuesEvent) {
            not = packQueryAllValuesEventReply ((QueryAllValuesEvent )event);
        }
        else if (event instanceof QueryGaugeStateEvent) {
            not = packQueryGaugeStateReply ((QueryGaugeStateEvent )event);
        }
        else if (event instanceof FindGaugeMgrEvent) {
            not = packFindGaugeMgrReply ((FindGaugeMgrEvent )event);
        }
        else if (event instanceof GaugeInitEvent) {
            not = packGaugeInitReply ((GaugeInitEvent )event);
        }
        else if (event instanceof GMgrDeleteGaugeEvent) {
            not = packGMgrDeleteGaugeReply ((GMgrDeleteGaugeEvent )event);
            Global.debug ("Publishing delete: (no direction or initiator) " + not.toString ());
        }
        
        // The event will be a response, so need to add the initiator
        // to make sure it gets back to the right requestor
        not.putAttribute (DIRECTION, RESPONSE);
        not.putAttribute (INITIATOR, initiator);
        
        // Publish the event using Siena
        try {
            Global.debug (not.toString ());
            siena.publish (not);
        }
        catch (SienaException ex) {
            ex.printStackTrace ();
            return false;
        }
        return true;
    }
    
    private NotificationBuffer prepareForReply (GaugeEvent event, Filter f) {
        
        // Only add gauge info for the right events
        // The following gauge events go to a gauge manager
        // so it doesn't make sense to add any gauge info
        if (!(event instanceof CreateGaugeEvent) && 
            !(event instanceof QueryMetaInfoEvent) &&
            !(event instanceof FindGaugeMgrEvent ) &&
            !(event instanceof GMgrDeleteGaugeEvent) &&
            !(event instanceof QueryGaugeStateEvent)) {
            addGaugeInfo (event, f);
        }
        
        // Determine the right event to wait for and construct
        // a filter (pattern) to wait for on the bus
        if (event instanceof ConfigureGaugeEvent) {
            f.addConstraint (EVENTTYPE, CONFIGURE);
        }
        else if (event instanceof GaugeInitEvent) {
            f.addConstraint (EVENTTYPE, GAUGEINIT);
            f.addConstraint (GAUGEMGRID, ((GaugeInitEvent )event).gaugeMgrID);
        }
        else if (event instanceof DeleteGaugeEvent) {
            f.addConstraint (EVENTTYPE, DELETE);
            f.addConstraint (GAUGEMGRID, ((DeleteGaugeEvent )event).gaugeMgrID);
        }
        else if (event instanceof QueryValueEvent) {
            f.addConstraint (EVENTTYPE, QUERYVALUE);
        }
        else if (event instanceof QueryAllValuesEvent) {
            f.addConstraint (EVENTTYPE, QUERYALLVALUES);
        }
        else if (event instanceof QueryGaugeStateEvent) {
            f.addConstraint (EVENTTYPE, QUERYSTATE);
        }
        else if (event instanceof CreateGaugeEvent) {
            f.addConstraint (EVENTTYPE, CREATE);
            f.addConstraint (GAUGEMGRID, ((CreateGaugeEvent )event).gaugeMgrID);
            f.addConstraint (EVENTID, ((CreateGaugeEvent )event).createID);
        }
        else if (event instanceof QueryMetaInfoEvent) {
            f.addConstraint (EVENTTYPE, METAINFO);
            f.addConstraint (GAUGEMGRID, ((QueryMetaInfoEvent )event).gaugeMgrID);
            f.addConstraint (EVENTID, ((QueryMetaInfoEvent )event).qmiID);
        }
        else if (event instanceof FindGaugeMgrEvent) {
            f.addConstraint (EVENTTYPE, FINDGMGR);
            f.addConstraint (EVENTID, ((FindGaugeMgrEvent )event).findID);
        }
        else if (event instanceof GMgrDeleteGaugeEvent) {
 
            f.addConstraint (EVENTTYPE, GMGRDELETEGAUGE);
            f.addConstraint (GAUGEMGRID, ((GMgrDeleteGaugeEvent )event).gaugeMgrID);
            f.addConstraint (EVENTID, ((GMgrDeleteGaugeEvent )event).deleteID);

        }
        
        // Only wait for responses that I initiated
        f.addConstraint (INITIATOR, id);
        f.addConstraint (DIRECTION, RESPONSE);
        
        if (event instanceof QueryGaugeStateEvent) {
            System.err.println ("Waiting for: " + f.toString ());
        }
        
        NotificationBuffer nb = new NotificationBuffer ();
        try {
            siena.subscribe (f, nb);
        }
        catch (Exception ex) {
            ex.printStackTrace ();
            return null;
        }
        
        return nb;
        
    }
    
    /** Blocks, waiting for a reply, and fills in the reply event. This
     * should follow a publishRequest call.
     * @return The event containing the reply
     * @param nb The notification buffer created for synchronous
     * event communication.
     * @param f The filter containing the pattern of the desired reply.
     * @param event The event that will contain the reply
     */    
    private GaugeEvent waitForReply (GaugeEvent event, NotificationBuffer nb, Filter f) {

        Global.debug ("Waiting for reply: " + f.toString ());
        // Siena's blocking wait
        Notification reply = null;
        try {

            for (int i = 0; i < RETRIES && reply == null;  i++) {
  
                reply = nb.getNotification (TIMEOUT);
  
            }
            
            if (event instanceof QueryGaugeStateEvent) {
                System.out.println ("QGS");
            }
            
            Global.debug (reply==null?"null":reply.toString ());

           
        }
        catch (Exception ex) {
            ex.printStackTrace ();
        }
        
        if (null == reply) {
            event.status = false;
            return event;
        }
       
        // Once we have the reply, and it is successful, we need to 
        // turn the notification into the right reply event
        if (reply.getAttribute (STATUS).stringValue ().equals ("Ok")) {
            if (event instanceof ConfigureGaugeEvent) {
                unpackConfigureEventReply (reply, (ConfigureGaugeEvent )event);
            }
            else if (event instanceof DeleteGaugeEvent) {
                unpackDeleteEventReply (reply, (DeleteGaugeEvent )event);
            }
            else if (event instanceof QueryValueEvent) {
                unpackQueryValueEventReply (reply, (QueryValueEvent )event);
            }
            else if (event instanceof QueryAllValuesEvent) {
                unpackQueryAllValuesEventReply(reply, (QueryAllValuesEvent )event);
            }
            else if (event instanceof QueryGaugeStateEvent) {
                unpackQueryGaugeStateReply (reply, (QueryGaugeStateEvent )event);
            }
            else if (event instanceof CreateGaugeEvent) {
                unpackCreateEventReply (reply, (CreateGaugeEvent )event);
            }
            else if (event instanceof QueryMetaInfoEvent) {
                unpackQueryMetaInfoEventReply (reply, (QueryMetaInfoEvent )event);
            }
            else if (event instanceof GaugeInitEvent) {
                unpackGaugeInitEventReply (reply, (GaugeInitEvent )event);
            }
            else if (event instanceof FindGaugeMgrEvent) {
                unpackFindGaugeMgrReply (reply, (FindGaugeMgrEvent )event);
            }
            else if (event instanceof GMgrDeleteGaugeEvent) {
                unpackGMgrDeleteGaugeReply (reply, (GMgrDeleteGaugeEvent )event);
            }
            else {
                event.status = false;
            }
        }
        
        // Make  sure we don't get notified of other extraneous events
//        siena.unsubscribe (f, nb);
        return event;
    }
    
    /** Flattens all the gauge info to the notification. The notification will have the
     * attribute, value pairs:<br>
     * ("gaugeID", gauge.uid)<br>
     * ("gaugeName", gauge.gaugeName)<br>
     * ("gaugeType", gauge.gaugeType)<br>
     * ("modelName", gauge.modelName)<br>
     * ("modelType", gauge.modelType)
     * @param gauge The ID of the gauge to turn into a notification
     * @param n The notification to contain the flattened gauge ID
     */    
    public static void addGaugeInfo (GaugeID gauge, Notification n) {
        n.putAttribute ("gaugeID", gauge.uid);
        n.putAttribute ("gaugeType", gauge.gaugeType);
        n.putAttribute ("gaugeName", gauge.gaugeName);
        n.putAttribute ("modelType", gauge.modelType);
        n.putAttribute ("modelName", gauge.modelName);
    }
    
    private void extractGaugeID (GaugeID gauge, Notification n) {
        AttributeValue gv = n.getAttribute ("gaugeID");
        if (gv != null) {
            gauge.uid = gv.stringValue ();
        }
        gauge.gaugeType = n.getAttribute ("gaugeType").stringValue ();
        gauge.gaugeName = n.getAttribute ("gaugeName").stringValue ();
        gauge.modelType = n.getAttribute ("modelType").stringValue ();
        gauge.modelName = n.getAttribute ("modelName").stringValue ();
    }
    
    private Notification packConfiguredEvent(ConfiguredEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, CONFIGURED);
        addGaugeInfo (event, n);
        n.putAttribute (STATUS, event.status?"Ok":"Err");
        this.SPV2Notification (event.configParams, n, CONFIG);
        return n;
    }
    
    private ConfiguredEvent unpackConfiguredEvent (Notification e) {
        GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, e);
        ConfiguredEvent event = new ConfiguredEvent (gaugeID);
        event.status = e.getAttribute (STATUS).stringValue ().equals ("Ok");
        if (event.status) {
            event.configParams = new StringPairVector ();
            Notification2SPV (e, event.configParams, CONFIG);
        }
        return event;
    }
    
    private Notification packCreatedEvent(CreatedEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, CREATED);
        addGaugeInfo (event, n);
        n.putAttribute (STATUS, event.status?"Ok":"Err");
        return n;
    }
    
    private CreatedEvent unpackCreatedEvent (Notification e) {
        GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, e);
        CreatedEvent event = new CreatedEvent (gaugeID);
        event.status = e.getAttribute (STATUS).stringValue ().equals ("Ok");
        return event;
    }
    
    private Notification packDeletedEvent(DeletedEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, DELETED);
        addGaugeInfo (event, n);
        n.putAttribute (GAUGEMGRID, event.gaugeMgrID);
        n.putAttribute (STATUS, event.status?"Ok":"Err");
        return n;
    }
    
    private DeletedEvent unpackDeletedEvent (Notification e) {
        GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, e);
        DeletedEvent event = new DeletedEvent (gaugeID);
        event.gaugeMgrID = e.getAttribute (GAUGEMGRID).stringValue ();
        event.status = e.getAttribute (STATUS).stringValue ().equals ("Ok");
        return event;
    }
    
    private Notification packMultipleValueEvent(MultipleValueEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, MULTIPLEVALUES);
        addGaugeInfo (event, n);
        this.GVV2Notification (event.values, n, "value");
        return n;
    }
    
    private MultipleValueEvent unpackMultipleValueEvent (Notification n) {
        GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, n);
        MultipleValueEvent event = new MultipleValueEvent (gaugeID);
        event.values = new GaugeValueVector ();
        Notification2GVV (n, event.values, "value");
        return event;
    }
    
    private Notification packValueEvent(ValueEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, VALUE);
        addGaugeInfo (event, n);
        n.putAttribute ("valueName", event.valueName);
        n.putAttribute ("propertyName", event.propertyName);
        n.putAttribute ("value", event.value);
        return n;
    }
    
    private ValueEvent unpackValueEvent (Notification n) {
        GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, n);
        ValueEvent event = new ValueEvent (gaugeID);
        event.valueName = n.getAttribute ("valueName").stringValue ();
        event.propertyName = n.getAttribute ("propertyName").stringValue ();
        event.value = n.getAttribute ("value").stringValue ();
        return event;
    }
    
    private void unpackQueryGaugeStateReply (Notification n, QueryGaugeStateEvent event) {
        event.status = n.getAttribute (STATUS).stringValue ().equals ("Ok");
        if (event.status) {
            event.setupParams = new StringPairVector ();
            event.configParams = new StringPairVector ();
            event.mappings = new StringPairVector ();
            this.Notification2SPV (n, event.setupParams, SETUP);
            this.Notification2SPV (n, event.configParams, CONFIG);
            this.Notification2SPV (n, event.mappings, MAPPINGS);
        }
    }
    
    private QueryGaugeStateEvent unpackQueryGaugeStateRequest (Notification n) {
        GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, n);
        return new QueryGaugeStateEvent (gaugeID);
    }
    
    private Notification packQueryGaugeStateReply (QueryGaugeStateEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, QUERYSTATE);
        n.putAttribute (STATUS, event.status?"Ok":"Err");
        addGaugeInfo (event, n);
        this.SPV2Notification (event.setupParams, n, SETUP);
        this.SPV2Notification (event.configParams, n, CONFIG);
        this.SPV2Notification (event.mappings, n, MAPPINGS);
        return n;
    }
    
    private Notification packQueryGaugeStateRequest (QueryGaugeStateEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, QUERYSTATE);
        addGaugeInfo (event, n);
        return n;
    }
    
    /** Flattens all the gauge info to the filter. The notification will have the
     * attribute, value pairs:<br>
     * ("gaugeID", gauge.uid)<br>
     * ("gaugeName", gauge.gaugeName)<br>
     * ("gaugeType", gauge.gaugeType)<br>
     * ("modelName", gauge.modelName)<br>
     * ("modelType", gauge.modelType)
     * @param event The gauge ID to be flattened.
     * @param f The filter to contain the flattened GaugeID
     */    
    public static void addGaugeInfo(GaugeID event, Filter f) {
        f.addConstraint ("gaugeType", event.gaugeType);
        f.addConstraint ("gaugeName", event.gaugeName);
        f.addConstraint ("modelType", event.modelType);
        f.addConstraint ("modelName", event.modelName);
    }
    
    private Notification packQueryMetaInfoEventRequest(QueryMetaInfoEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, METAINFO); 
        n.putAttribute (GAUGEMGRID, event.gaugeMgrID);
        n.putAttribute (EVENTID, event.qmiID);
        n.putAttribute ("gaugeType", event.gaugeType);
        return n;
    }
    
    private Notification packQueryMetaInfoEventReply (QueryMetaInfoEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, METAINFO);
        n.putAttribute (GAUGEMGRID, event.gaugeMgrID);
        n.putAttribute ("gaugeType", event.gaugeType);
        n.putAttribute (EVENTID, event.qmiID);
        this.SPV2Notification (event.configParams, n, CONFIG);
        this.SPV2Notification (event.values, n, "value");
        n.putAttribute (STATUS, event.status?"Ok":"Err");
        return n;
    }
    
    private void unpackQueryMetaInfoEventReply (Notification n, QueryMetaInfoEvent event) {
        event.status = n.getAttribute (STATUS).stringValue ().equals ("Ok");
        if (event.status) {
            event.gaugeType = n.getAttribute ("gaugeType").stringValue ();
            event.gaugeMgrID = n.getAttribute (GAUGEMGRID).stringValue ();
            event.qmiID = n.getAttribute (EVENTID).stringValue ();
            this.Notification2SPV (n, event.configParams, CONFIG);
            this.Notification2SPV (n, event.values, "value");
        }
    }
    
    private QueryMetaInfoEvent unpackQueryMetaInfoEventRequest (Notification n) {
         String gaugeType = n.getAttribute ("gaugeType").stringValue ();
         QueryMetaInfoEvent event = new QueryMetaInfoEvent (gaugeType);
        event.qmiID = n.getAttribute (EVENTID).stringValue ();
        return event;
    }
    
    private Notification packCreateEventRequest (CreateGaugeEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, CREATE);
        n.putAttribute (GAUGEMGRID, event.gaugeMgrID);
        n.putAttribute (EVENTID, event.createID);
        addGaugeInfo (event, n);
        this.SPV2Notification (event.setupParams, n, SETUP);
        this.SPV2Notification (event.mappings, n, MAPPINGS);
        return n;
    }
    
    private void unpackCreateEventReply (Notification n, CreateGaugeEvent event) {
        event.status = n.getAttribute (STATUS).stringValue ().equals ("Ok");
        event.gaugeMgrID = n.getAttribute (GAUGEMGRID).stringValue ();
        event.createID = n.getAttribute (EVENTID).stringValue ();
        this.extractGaugeID (event, n);
    }
    
    private CreateGaugeEvent unpackCreateEventRequest (Notification n) {
        GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, n);
        CreateGaugeEvent event = new CreateGaugeEvent (gaugeID);
        event.gaugeMgrID = n.getAttribute (GAUGEMGRID).stringValue ();
        event.createID = n.getAttribute (EVENTID).stringValue ();
        event.setupParams = new StringPairVector ();
        event.mappings = new StringPairVector ();
        this.Notification2SPV (n, event.setupParams, SETUP);
        this.Notification2SPV (n, event.mappings, MAPPINGS);
        return event;
    }
  
    private Notification packCreateEventReply (CreateGaugeEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, CREATE);
        n.putAttribute (GAUGEMGRID, event.gaugeMgrID);
        n.putAttribute (EVENTID, event.createID);
        addGaugeInfo (event, n);
        n.putAttribute (STATUS, event.status?"Ok":"Err");
        return n;
    }
    
    private Notification packConfigureEventRequest(ConfigureGaugeEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, CONFIGURE);
        addGaugeInfo (event, n);
        this.SPV2Notification (event.configParams, n, CONFIG);
        return n;
    }
    
    private void unpackConfigureEventReply (Notification n, ConfigureGaugeEvent event) {
        event.status = n.getAttribute (STATUS).stringValue ().equals ("Ok");
        this.extractGaugeID (event, n);
    }
    
    private ConfigureGaugeEvent unpackConfigureGaugeRequest (Notification n) {
        GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, n);
        ConfigureGaugeEvent event = new ConfigureGaugeEvent (gaugeID);
        event.configParams = new StringPairVector ();
        this.Notification2SPV (n, event.configParams, CONFIG);
        return event;
    }   
    
    private Notification packConfigureEventReply (ConfigureGaugeEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, CONFIGURE);
        addGaugeInfo (event, n);
        n.putAttribute (STATUS, event.status?"Ok":"Err");
        return n;
    }
    
    private void unpackDeleteEventReply (Notification n, DeleteGaugeEvent event) {
        event.status = n.getAttribute (STATUS).stringValue ().equals ("Ok");
        event.gaugeMgrID = n.getAttribute (GAUGEMGRID).stringValue ();
        this.extractGaugeID (event, n);
    }
    
    private DeleteGaugeEvent unpackDeleteEventRequest (Notification n) {
        GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, n);
        DeleteGaugeEvent event = new DeleteGaugeEvent (gaugeID);
        event.gaugeMgrID = n.getAttribute (GAUGEMGRID).stringValue ();
        return event;
    }
    
    private Notification packDeleteEventRequest(DeleteGaugeEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, DELETE);
        n.putAttribute (GAUGEMGRID, event.gaugeMgrID);
        addGaugeInfo (event, n);
        return n;
    }
    
    private Notification packDeleteEventReply (DeleteGaugeEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, DELETE);
        n.putAttribute (GAUGEMGRID, event.gaugeMgrID);
        addGaugeInfo (event, n);
        n.putAttribute (STATUS, event.status?"Ok":"Err");
        return n;
    }
    
    private Notification packQueryValueEventRequest(QueryValueEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, QUERYVALUE);
        n.putAttribute ("valueName", event.valueName);
        addGaugeInfo (event, n);
        return n;
    }
    
    private QueryValueEvent unpackQueryValueEventRequest (Notification n) {
        GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, n);
        QueryValueEvent event = new QueryValueEvent (gaugeID);
        event.valueName = n.getAttribute ("valueName").stringValue ();
        return event;
    }
    
    private void unpackQueryValueEventReply (Notification n, QueryValueEvent event) {
        event.status = n.getAttribute (STATUS).stringValue ().equals ("Ok");
        if (event.status) {
            event.valueName = n.getAttribute ("valueName").stringValue ();
            event.value = n.getAttribute ("value").stringValue ();
        }
    }
    
    private Notification packQueryValueEventReply (QueryValueEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, QUERYVALUE);
        addGaugeInfo (event, n);
        n.putAttribute ("valueName", event.valueName);
        n.putAttribute ("value", event.value);
        n.putAttribute ("status", event.status?"Ok":"Err");
        return n;
    }
    
    private Notification packQueryAllValuesEventRequest(QueryAllValuesEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, QUERYALLVALUES);
        addGaugeInfo (event, n);
        return n;
    }
    
    private QueryAllValuesEvent unpackQueryAllValuesEventRequest (Notification n) {
        GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, n);
        QueryAllValuesEvent event = new QueryAllValuesEvent (gaugeID);
        return event;
    }
    
    private void unpackQueryAllValuesEventReply (Notification n, QueryAllValuesEvent event) {
        event.status = n.getAttribute (STATUS).stringValue ().equals ("Ok");
        if (event.status) {
            event.values = new GaugeValueVector ();
            this.Notification2GVV (n, event.values, "value");
        }
    }
    
    private Notification packQueryAllValuesEventReply (QueryAllValuesEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, QUERYALLVALUES);
        addGaugeInfo (event, n);
        this.GVV2Notification (event.values, n, "value");
        n.putAttribute ("status", event.status?"Ok":"Err");
        return n;
    }
    
    private Notification packFindGaugeMgrRequest (FindGaugeMgrEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, FINDGMGR);
        n.putAttribute ("gaugeType", event.gaugeType);
        n.putAttribute (EVENTID, event.findID);
        return n;
    }
    
    private Notification packFindGaugeMgrReply (FindGaugeMgrEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, FINDGMGR);
        n.putAttribute ("gaugeType", event.gaugeType);
        n.putAttribute (EVENTID, event.findID);
        n.putAttribute (STATUS, event.status?"Ok":"Err");
        n.putAttribute (GAUGEMGRID, event.gaugeMgrID);
        return n;
    }
    
    private void unpackFindGaugeMgrReply (Notification n, FindGaugeMgrEvent event) {
        event.status = n.getAttribute (STATUS).stringValue ().equals ("Ok");
        if (event.status) {
            event.gaugeMgrID = n.getAttribute (GAUGEMGRID).stringValue ();
            event.findID = n.getAttribute ("eventID").stringValue ();
        }
    }
    
    private FindGaugeMgrEvent unpackFindGaugeMgrRequest (Notification n) {
        String gaugeType = n.getAttribute ("gaugeType").stringValue ();
        FindGaugeMgrEvent event = new FindGaugeMgrEvent (gaugeType);
        event.findID = n.getAttribute ("eventID").stringValue ();
        return event;
    }
    
    private Notification packGMgrDeleteGaugeRequest (GMgrDeleteGaugeEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, GMGRDELETEGAUGE);
        addGaugeInfo (event, n);
        n.putAttribute (GAUGEMGRID, event.gaugeMgrID);
        n.putAttribute (EVENTID, event.deleteID);
        return n;
    }
    
    private Notification packGMgrDeleteGaugeReply (GMgrDeleteGaugeEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, GMGRDELETEGAUGE);
        addGaugeInfo (event, n);
        n.putAttribute (GAUGEMGRID, event.gaugeMgrID);
        n.putAttribute (EVENTID, event.deleteID);
        n.putAttribute (STATUS, event.status?"Ok":"Err");
        return n;
    }
    
    private GMgrDeleteGaugeEvent unpackGMgrDeleteGaugeRequest(Notification n) {
       GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, n);
        GMgrDeleteGaugeEvent event = new GMgrDeleteGaugeEvent (gaugeID);
        event.gaugeMgrID = n.getAttribute(GAUGEMGRID).stringValue();
        event.deleteID = n.getAttribute(EVENTID).stringValue ();
        return event;
        
    }
    
    private void unpackGMgrDeleteGaugeReply (Notification n, GMgrDeleteGaugeEvent event) {
 
        event.status = n.getAttribute (STATUS).stringValue ().equals ("Ok");
        if (event.status) {
            extractGaugeID (event, n);
            event.gaugeMgrID = n.getAttribute (GAUGEMGRID).stringValue ();
            event.deleteID = n.getAttribute (EVENTID).stringValue ();
        }
    }
    
    private Notification packGaugeInitRequest (GaugeInitEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, GAUGEINIT);
        addGaugeInfo (event, n);
        n.putAttribute (GAUGEMGRID, event.gaugeMgrID);
        return n;
    }
    
    private Notification packGaugeInitReply (GaugeInitEvent event) {
        Notification n = new Notification ();
        addGaugeInfo (event, n);
        n.putAttribute (EVENTTYPE, GAUGEINIT);
        n.putAttribute (GAUGEMGRID, event.gaugeMgrID);
        this.SPV2Notification (event.setup, n, SETUP);
        this.SPV2Notification (event.mappings, n, MAPPINGS);
        n.putAttribute (STATUS, event.status?"Ok":"Err");
        return n;
    }
    
    private void unpackGaugeInitEventReply (Notification n, GaugeInitEvent event) {
        event.gaugeMgrID = n.getAttribute (GAUGEMGRID).stringValue ();
        event.status = n.getAttribute (STATUS).stringValue ().equals ("Ok");
        if (event.status) {
            this.Notification2SPV (n, event.setup, SETUP);
            this.Notification2SPV (n, event.mappings, MAPPINGS);
        }
       
    }
    
    private GaugeInitEvent unpackGaugeInitEventRequest (Notification n) {
        GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, n);
        GaugeInitEvent event = new GaugeInitEvent (gaugeID);
        event.gaugeMgrID = n.getAttribute (GAUGEMGRID).stringValue ();
        return event;
    }
    
    private Notification packCreationFinishedRequest (CreationFinishedEvent event) {
        Notification n = new Notification ();
        n.putAttribute (EVENTTYPE, FINISHEDCREATION);
        addGaugeInfo (event, n);
        n.putAttribute (STATUS, event.status?"Ok":"Failed");
        n.putAttribute (GAUGEMGRID, event.gaugeMgrID);
        return n;
    }
    
    private CreationFinishedEvent unpackCreationFinishedRequest (Notification n) {
        GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, n);
        CreationFinishedEvent event = new CreationFinishedEvent (gaugeID);
        event.gaugeMgrID = n.getAttribute (GAUGEMGRID).stringValue ();
        event.status = n.getAttribute (STATUS).stringValue ().equals ("Ok");
        return event;
    }
    
    private void SPV2Notification (StringPairVector spv, Notification e, String prefix) {
        if (null == e) return;
        
        Integer size = new Integer (spv.size ());
        e.putAttribute ("vectorSize_" + prefix, size.toString ());
        for (int i = 0; i < spv.size (); i++) {
            e.putAttribute ("name_" + prefix + "_" + i, spv.nameAt (i));
            e.putAttribute ("value_" + prefix + "_" + i, spv.valueAt (i));
        }
    }
    
    private void Notification2SPV (Notification e, StringPairVector spv, String prefix) {
        String pattern = "vectorSize_" + prefix;
        AttributeValue av = e.getAttribute (pattern);
        int size = av.intValue ();
        for (int i = 0; i < size; i++) {
            spv.addElement (e.getAttribute ("name_" + prefix + "_" + i).stringValue (),
                            e.getAttribute ("value_" + prefix + "_" + i).stringValue ());
        }
    }
    
    private void GVV2Notification (GaugeValueVector gvv, Notification e, String prefix) {
        if (null == e) return;
        
        Integer size = new Integer (gvv.size ());
        e.putAttribute ("vectorSize_" + prefix, size.toString ());
        for (int i = 0; i < gvv.size (); i++) {
            e.putAttribute ("valueName_" + prefix + "_" + i, gvv.nameAt (i));
            e.putAttribute ("propertyName_" + prefix + "_" + i, gvv.propertyAt (i));
            e.putAttribute ("value_" + prefix + "_" + i, gvv.valueAt (i));
        } 
    }
    
    private void Notification2GVV (Notification e, GaugeValueVector gvv, String prefix) {
        String pattern = "vectorSize_" + prefix;
        int size = e.getAttribute (pattern).intValue ();
        for (int i = 0; i < size; i++) {
            gvv.addElement (e.getAttribute ("valueName_" + prefix + "_" + i).stringValue (),
                            e.getAttribute ("propertyName_" + prefix + "_" + i).stringValue (),
                            e.getAttribute ("value_" + prefix + "_" + i).stringValue ());
        }
    }
    
    // Only handles GaugeConsumer subscription at the moment
    /** Notifies GaugeConsumers of an event that is of interest to them
     * @param e The event
     */    
    public void notifyGaugeConsumer (Notification e) {
        GaugeID gauge = new GaugeID ();
        extractGaugeID (gauge, e);       
        boolean matches = false;
        Enumeration bindings = eventBindings.elements ();
        
        while (bindings.hasMoreElements ()) {
           NotificationHandlerGauge fh = (NotificationHandlerGauge )bindings.nextElement ();
            if (gauge.equals (fh.gauge.getGaugeID ())) {
                // Get event type
                AttributeValue av = fh.filterAsNotification.getAttribute (EVENTTYPE);
                if (av != null && e.getAttribute (EVENTTYPE).stringValue ().equals (av.stringValue ())) {
                    String eventType = av.stringValue ();
                    if (eventType.equals (CREATED)) {
                        if (fh.handler instanceof GaugeConsumer) {
                            ((GaugeConsumer )fh.handler).onReportCreated (fh.gauge);
                        }
                    }
                    else if (eventType.equals (DELETED)) {
                        if (fh.handler instanceof GaugeConsumer) {
                            ((GaugeConsumer )fh.handler).onReportDeleted (fh.gauge);
                        }
                    }
                    else if (eventType.equals (CONFIGURED)) {
                        if (fh.handler instanceof GaugeConsumer) {
                            StringPairVector configParams = new StringPairVector ();
                            Notification2SPV (e, configParams, CONFIG);
                            ((GaugeConsumer )fh.handler).onReportConfigured (fh.gauge, configParams);
                        }
                    }
                    else if (eventType.equals (VALUE)) {
                        if (fh.handler instanceof GaugeConsumer) {
                            String valueName = e.getAttribute ("valueName").stringValue ();
                            String propertyName = e.getAttribute ("propertyName").stringValue ();
                            String value = e.getAttribute ("value").stringValue ();
                            ((GaugeConsumer )fh.handler).onReportValue (fh.gauge, valueName, propertyName, value);
                        }
                    }
                    else if (eventType.equals (MULTIPLEVALUES)) {
                        if (fh.handler instanceof GaugeConsumer) {
                            GaugeValueVector gvv = new GaugeValueVector ();
                            Notification2GVV (e, gvv, "value");
                            ((GaugeConsumer )fh.handler).onReportMultipleValues (fh.gauge, gvv);
                        }
                    }
                }
            }
        }
    }
    
    /** Notifies a gauge of events of interest. Called by Siena when
     * an event matching a filter is matched.
     * @param e The event
     */    
    public void notifyGauge (Notification e) {
        GaugeID gaugeID = new GaugeID ();
        extractGaugeID (gaugeID, e);
        GaugeControl gauge = null;
        Enumeration gauges = this.gauges.elements ();
        while (gauges.hasMoreElements () && gauge == null) {
            GaugeControl candidate = (GaugeControl )gauges.nextElement ();
            if (candidate.getGaugeID ().equals (gaugeID)) {
                gauge = candidate;
            }
        }
        
        if (gauge != null) {
            String eventType = e.getAttribute (EVENTTYPE).stringValue ();
            String initiator = e.getAttribute (INITIATOR).stringValue ();
            if (eventType.equals (CONFIGURE)) {
                ConfigureGaugeEvent reply = unpackConfigureGaugeRequest (e);
                
                reply.status = gauge.configure (reply.configParams);
                
                // Announce that the gauge has been configured
                ConfiguredEvent announce = new ConfiguredEvent (gaugeID);
                announce.configParams = reply.configParams;
                announce.status = reply.status;
                publishAnnouncement (announce);
                
                // Reply to sender
                publishResponse (reply, initiator);
            }
            else if (eventType.equals (QUERYSTATE)) {
                QueryGaugeStateEvent reply = unpackQueryGaugeStateRequest (e);
                reply.setupParams = new StringPairVector ();
                reply.configParams = new StringPairVector ();
                reply.mappings = new StringPairVector ();
                reply.status = gauge.queryState (reply.setupParams, reply.configParams, reply.mappings);
                publishResponse (reply, initiator);
            }
            else if (eventType.equals (QUERYVALUE)) {
                QueryValueEvent reply = unpackQueryValueEventRequest (e);
                reply.value = gauge.queryValue (reply.valueName);
                if (reply.value != null && !reply.value.equals ("")) reply.status = true;
                else reply.status = false;
                publishResponse (reply, initiator);
            }
            else if (eventType.equals (QUERYALLVALUES)) {
                QueryAllValuesEvent reply = unpackQueryAllValuesEventRequest (e);
                reply.values = new GaugeValueVector ();
                reply.status = gauge.queryAllValues (reply.values);
                publishResponse (reply, initiator);
            }
            else if (eventType.equals (GMGRDELETEGAUGE)) {
                GMgrDeleteGaugeEvent reply = unpackGMgrDeleteGaugeRequest (e);
                reply.status = true;
                ((SienaGauge )gauge).processDelete (reply.gaugeMgrID);
                publishResponse (reply, initiator);
            }
        }
    }
    
    private void notifyGaugeMgr (Notification e) {
        // Handles:
        // findGaugeMgrEvent (handled internally by bus)
        // GaugeInitEvents (sent by a gauge after the process has been created)
        // CreateGaugeEvent, DeleteGaugeEvent, QueryMetaInfoEvent, which are processed by the gauge mgr

        if (!gaugeMgrs.isEmpty ()) {

            String eventType = e.getAttribute (EVENTTYPE).stringValue ();
      
            if (eventType.equals (FINDGMGR)) {
                String initiator = e.getAttribute (INITIATOR).stringValue ();
                FindGaugeMgrEvent event = unpackFindGaugeMgrRequest (e);
                Enumeration mgrs = gaugeMgrs.elements ();
                boolean foundGM = false;
                while (mgrs.hasMoreElements () && !foundGM) {
                    GaugeMgr gaugeMgr = (GaugeMgr )mgrs.nextElement ();
                    if (gaugeMgr.managesType (event.gaugeType)) {
                        foundGM = true;
                        event.gaugeMgrID = gaugeMgr.getGaugeMgrID ();
                        event.status = true;
                        publishResponse (event, initiator);
                    }
                }
            }
            else if (eventType.equals (GAUGEINIT)) {
                GaugeInitEvent init = unpackGaugeInitEventRequest (e);
                String initiator = e.getAttribute (INITIATOR).stringValue ();
                // find the matching creation event
                CreateGaugeEvent event = null;
                for (int i = 0; i < bufferedCreationEvents.size () && event == null; i++) {
                    CreateGaugeEvent candidate = (CreateGaugeEvent )bufferedCreationEvents.elementAt (i);
                    if (candidate.equals (init)) {
                        event = candidate;
                    }
                }
                // Send gauge setups and mappings
                init.setup = event.setupParams;
                init.mappings = event.mappings;
                init.status = true;
                publishResponse (init, initiator);
            }
            else if (eventType.equals (FINISHEDCREATION)) {
                CreationFinishedEvent fin = unpackCreationFinishedRequest (e);
                CreateGaugeEvent event = null;
                for (int i = 0; i < bufferedCreationEvents.size () && event == null; i++) {
                    CreateGaugeEvent candidate = (CreateGaugeEvent )bufferedCreationEvents.elementAt (i);
                    if (candidate.equals (fin)) {
                        event = candidate;
                    }
                }
                event.status = fin.status;
                String consumer = event.gaugeMgrID;
                event.gaugeMgrID = fin.gaugeMgrID;
                publishResponse (event, consumer);
            }
            else if (eventType.equals (CREATE)) {
                String initiator = e.getAttribute (INITIATOR).stringValue ();             
                CreateGaugeEvent event = unpackCreateEventRequest (e);
                Enumeration mgrs = gaugeMgrs.elements ();
                boolean foundGM = false;
                while (mgrs.hasMoreElements () && !foundGM) {
                    GaugeMgr gaugeMgr = (GaugeMgr )mgrs.nextElement ();

                     if (gaugeMgr.getGaugeMgrID ().equals (event.gaugeMgrID)) {
                        foundGM = true;
                        GaugeControl gauge = gaugeMgr.createGauge (event, event.setupParams, event.mappings);
                        event.gaugeMgrID = initiator; // hack to remember initiator
                        bufferedCreationEvents.addElement (event);
                        // The creation is finished in the GaugeInit event handler
                    }
                }
            }
            else if (eventType.equals (DELETE)) {
                String initiator = e.getAttribute (INITIATOR).stringValue ();
                DeleteGaugeEvent event = unpackDeleteEventRequest (e);
                Enumeration mgrs = gaugeMgrs.elements ();
                boolean foundGM = false;
                while (mgrs.hasMoreElements () && !foundGM) {
                    GaugeMgr gaugeMgr = (GaugeMgr )mgrs.nextElement ();
                    if (gaugeMgr.getGaugeMgrID ().equals (event.gaugeMgrID)) {
                        SienaGaugeMgrGaugeHandle gauge = new SienaGaugeMgrGaugeHandle (event);
                        
                        // Need to tell the gauge under the covers
                        GMgrDeleteGaugeEvent deleteEvent = new GMgrDeleteGaugeEvent (gauge.getGaugeID ());
                        deleteEvent.gaugeMgrID = gaugeMgr.getGaugeMgrID ();
                        publishRequest (deleteEvent);
                        event.status = gaugeMgr.deleteGauge (gauge);
                        publishResponse (event, initiator);
                    }
                }
            }
            else if (eventType.equals (METAINFO)) {
                Global.debug ("Got MetaInfo");
                String initiator = e.getAttribute (INITIATOR).stringValue ();
                QueryMetaInfoEvent event = unpackQueryMetaInfoEventRequest (e);
                Enumeration mgrs = gaugeMgrs.elements ();
                boolean foundGM = false;
                while (mgrs.hasMoreElements () && !foundGM) {
                    Global.debug ("finding GM");
                    GaugeMgr gaugeMgr = (GaugeMgr )mgrs.nextElement ();
                    if (gaugeMgr.managesType (event.gaugeType)) {
                        Global.debug ("Found GM");
                        event.configParams = new StringPairVector ();
                        event.values = new StringPairVector ();
                        Global.debug ("Calling query for " + event.gaugeType);
                        event.gaugeMgrID = gaugeMgr.getGaugeMgrID ();
                        event.status = gaugeMgr.queryMetaInfo (event.gaugeType, event.configParams, event.values);
                        publishResponse (event, initiator);
                    }
                }
                
            }
            
        }
    }
    
    /** Called by Siena when a matching event is published. This works out whether
     * the event should be sent to a gauge consumer, gauge manager, or gauge.
     * @param e
     */    
    public void notify (Notification e) {
        Global.debug ("Notify: " + e);
        AttributeValue av = e.getAttribute (EVENTTYPE);
        if (av != null) {
            String eventType = av.stringValue ();
            AttributeValue av2 = e.getAttribute (DIRECTION);
            String direction = "";
            if (av2 != null) {
                direction = av2.stringValue ();
            }
            
            if (eventType.equals (MULTIPLEVALUES) || eventType.equals (VALUE) || eventType.equals (CONFIGURED) ||
                eventType.equals (CREATED) || eventType.equals (DELETED) || 
                // replies
                (direction.equals (RESPONSE) && (eventType.equals (CREATE) || eventType.equals (DELETE) ||
                                              eventType.equals (CONFIGURE) || eventType.equals (QUERYVALUE) ||
                                              eventType.equals (QUERYALLVALUES) || eventType.equals (QUERYSTATE) ||
                                              eventType.equals (FINDGMGR) ))) {
                notifyGaugeConsumer (e);
            }
            if (direction.equals (REQUEST) && (eventType.equals (CONFIGURE) || eventType.equals (QUERYSTATE) ||
                                                    eventType.equals (QUERYVALUE) || eventType.equals (QUERYALLVALUES) ||
                                                    eventType.equals (GMGRDELETEGAUGE))) {
                notifyGauge (e);
            }
            if ((direction.equals (REQUEST) && (eventType.equals (CREATE) || eventType.equals (DELETE) ||
                                                eventType.equals (METAINFO) || eventType.equals (GAUGEINIT) ||
                                                eventType.equals (FINDGMGR)))
                || eventType.equals (DELETED) || eventType.equals (FINISHEDCREATION)) {
                notifyGaugeMgr (e);
            }
        }
    }
    
    public void notify(Notification[] e) {
        for (int i = 0; i < e.length; i++) {
            notify (e [i]);
        }
    }
    
    private static Notification Filter2Notification (Filter f) {
        AttributeValue av;
        Notification n = new Notification ();
        
        Iterator it = f.constraintNamesIterator ();
        while (it.hasNext ()) {
            String name = (String )it.next ();
            Iterator it2 = f.constraintsIterator (name);
            n.putAttribute (name, ((AttributeConstraint )it2.next ()).value);
        }
        
        return n;
    }
    
    
    // This implements GaugeConsmumer event bus
    public boolean registerDeletedEvents (GaugeConsumer consumer, GaugeControl gauge) {
        return registerEvents (consumer, gauge, SienaEventBus.DELETED);
    }
    
    public boolean registerCreatedEvents (GaugeConsumer consumer, GaugeControl gauge) {
        return registerEvents (consumer, gauge, SienaEventBus.CREATED);
    }
    
    public boolean registerConfiguredEvents (GaugeConsumer consumer, GaugeControl gauge) {
        return registerEvents (consumer, gauge, SienaEventBus.CONFIGURED);
    }
    
    public boolean registerValueEvents (GaugeConsumer consumer, GaugeControl gauge) {
        return registerEvents (consumer, gauge, SienaEventBus.VALUE) &&
               registerEvents (consumer, gauge, SienaEventBus.MULTIPLEVALUES);
    }
    
    private boolean registerEvents (GaugeConsumer consumer, GaugeControl gauge, String eventType) {
        Filter f = new Filter ();
        SienaEventBus.addGaugeInfo (gauge.getGaugeID (), f);
        NotificationHandlerGauge fh = new NotificationHandlerGauge ();
        f.addConstraint (EVENTTYPE, eventType);
        fh.filterAsNotification = Filter2Notification (f);
        fh.filter = f;
        fh.handler = consumer;
        fh.gauge = gauge;
        eventBindings.addElement (fh);
        try {
            siena.subscribe (f, this);
        }
        catch (SienaException ex) {
            ex.printStackTrace ();
            return false;
        }
        return true;
    }
    
    public boolean unregisterGaugeEvents (GaugeConsumer consumer, GaugeControl gauge) {
        for (int i = 0; i < eventBindings.size (); i++) {
            NotificationHandlerGauge fh = (NotificationHandlerGauge )eventBindings.elementAt (i);
            if (fh.handler == consumer && fh.gauge == gauge) {
                eventBindings.removeElementAt (i);
                i--;
                siena.unsubscribe (fh.filter, this);
            }
        }
        return true;
    }
    
    // This implements Gauge side events
    
    
    public void registerGauge (GaugeControl gauge) {
        Global.debug ("Registering Gauge: " + gauge.getGaugeID ().toString ());
        gauges.addElement (gauge);
        registerGaugeSideEvents (gauge);
    }
    
    public void unregisterGauge (GaugeControl gauge) {
        Global.debug ("Unregistering gauge: " + gauge.getGaugeID ().toString ());
        gauges.removeElement (gauge);
        Filter f = new Filter ();
        GaugeID gaugeID = gauge.getGaugeID ();
        addGaugeInfo (gaugeID, f);
        if (gauge instanceof SienaGaugeHandle) {
            f.addConstraint (DIRECTION, ANNOUNCE);
        }
        else {
            f.addConstraint (DIRECTION, REQUEST);
        }
        siena.unsubscribe (f, this);
    }
      
    
    private void registerGaugeSideEvents (GaugeControl gauge) {
        Filter f = new Filter ();
        GaugeID gaugeID = gauge.getGaugeID ();
        addGaugeInfo (gaugeID, f);
        if (gauge instanceof SienaGaugeHandle) {
            f.addConstraint (DIRECTION, ANNOUNCE);
        }
        else {
            f.addConstraint (DIRECTION, REQUEST);
        }
       
  
        try {
            siena.subscribe (f, this);
        }
        catch (SienaException ex) {
            System.err.println (gauge.getGaugeID ().toString () + ": Error subscribing to Siena bus.");
            ex.printStackTrace ();
        }
    }
    
    private void addConstraint (Filter f, String tag, String pattern) {
        if (pattern != null && !pattern.equals ("")) {
            f.addConstraint (tag, pattern);
        }
    }
    
    private void addEventSpecificConstraints (Filter f, GaugeEvent event) {
        if (event instanceof ConfiguredEvent) {
            ConfiguredEvent ce = (ConfiguredEvent )event;
            addConstraint (f, EVENTTYPE, CONFIGURED);
            // TODO: config params
        }
        else if (event instanceof CreatedEvent) {
            CreatedEvent ce = (CreatedEvent )event;
            addConstraint (f, EVENTTYPE, CREATED);
            // TODO: setup and mappings
        }
        else if (event instanceof DeletedEvent) {
            DeletedEvent de = (DeletedEvent )event;
            addConstraint (f, EVENTTYPE, DELETED);
        }
        else if (event instanceof MultipleValueEvent) {
            MultipleValueEvent mve = (MultipleValueEvent )event;
            addConstraint (f, EVENTTYPE, MULTIPLEVALUES);
            // TODO: add values
        }
        else if (event instanceof ValueEvent) {
            ValueEvent ve = (ValueEvent )event;
            addConstraint (f, EVENTTYPE, VALUE);
            addConstraint (f, "valueName", ve.valueName);
            addConstraint (f, "propertyName", ve.propertyName);
            addConstraint (f, "value", ve.value);
        }
    }
    
    public Filter event2Filter(GaugeEvent event) {
        Filter f = new Filter();
        addConstraint(f, "gaugeID", event.uid);
        addConstraint(f, "gaugeName", event.gaugeName);
        addConstraint(f, "gaugeType", event.gaugeType);
        addConstraint(f, "modelName", event.modelName);
        addConstraint(f, "modelType", event.modelType);
        
        addEventSpecificConstraints(f, event);
        return f;
    }
    
    public GaugeEvent notification2Event (Notification e) {
        String eventType = e.getAttribute (EVENTTYPE).stringValue ();
        
        if (eventType.equals (CREATED)) {
            return unpackCreatedEvent (e);
        }
        else if (eventType.equals (CONFIGURED)) {
            return unpackConfiguredEvent (e);
        }
        else if (eventType.equals (VALUE)) {
            return unpackValueEvent (e);
        }
        else if (eventType.equals (MULTIPLEVALUES)) {
            return unpackMultipleValueEvent (e);
        }
        else if (eventType.equals (DELETED)) {
            return unpackDeletedEvent (e);
        }
        return null;
    }
    
    public void registerGaugeMgr (GaugeMgr gMgr) {
        Global.debug("Registering " + gMgr.getGaugeMgrID ());
        gaugeMgrs.addElement (gMgr);
        registerGaugeMgrSideEvents (gMgr);
    }
    
    Vector gaugeMgrs = new Vector ();
    Vector bufferedCreationEvents = new Vector ();
    Hashtable bufferedDeleteEvents = new Hashtable ();
    
    private void registerGaugeMgrSideEvents (GaugeMgr gMgr) {
        Filter cgf = new Filter ();
        Filter dgf = new Filter ();
        Filter gif = new Filter ();
        Filter fmf = new Filter ();
        Filter qmf = new Filter ();
        Filter df = new Filter ();
        Filter fcf = new Filter ();
        
        cgf.addConstraint (EVENTTYPE, CREATE);
        cgf.addConstraint (GAUGEMGRID, gMgr.getGaugeMgrID ());
        cgf.addConstraint (DIRECTION, REQUEST);
        
        dgf.addConstraint (EVENTTYPE, DELETE);
        dgf.addConstraint (GAUGEMGRID, gMgr.getGaugeMgrID ());
        dgf.addConstraint (DIRECTION, REQUEST);
         
        gif.addConstraint (EVENTTYPE, GAUGEINIT);
        gif.addConstraint (GAUGEMGRID, gMgr.getGaugeMgrID ());
        gif.addConstraint (DIRECTION, REQUEST);   
        
        fmf.addConstraint (EVENTTYPE, FINDGMGR);
        fmf.addConstraint (DIRECTION, REQUEST);
        
        qmf.addConstraint (EVENTTYPE, METAINFO);
        qmf.addConstraint (GAUGEMGRID, gMgr.getGaugeMgrID ());
        qmf.addConstraint (DIRECTION, REQUEST);   
        
        df.addConstraint (EVENTTYPE, DELETED);
        df.addConstraint (GAUGEMGRID, gMgr.getGaugeMgrID ());
        
        fcf.addConstraint (EVENTTYPE, FINISHEDCREATION);
        fcf.addConstraint (GAUGEMGRID, gMgr.getGaugeMgrID ());
        
        try {
            siena.subscribe (cgf, this);
            siena.subscribe (dgf, this);
            siena.subscribe (gif, this);
            siena.subscribe (fmf, this);
            siena.subscribe (qmf, this);
            siena.subscribe (df, this);
            siena.subscribe (fcf, this);
        }
        catch (SienaException ex) {
            System.err.println ("Gauge Manager: " + gMgr.getGaugeMgrID () + ": Error subscribing to Siena bus.");
            ex.printStackTrace ();
        }
    }
}
